"""
sphinx_exercise.directive
~~~~~~~~~~~~~~~~~~~~~~~~~

A custom Sphinx Directive

:copyright: Copyright 2020 by the QuantEcon team, see AUTHORS
:licences: see LICENSE for details
"""
from typing import List
from docutils.nodes import Node

from sphinx.util.docutils import SphinxDirective
from docutils.parsers.rst import directives
from .nodes import exercise_node, exercise_unenumerable_node, solution_node
from docutils import nodes
from sphinx.util import logging

logger = logging.getLogger(__name__)


class CustomDirective(SphinxDirective):
    """ A custom Sphinx directive """

    name = ""

    def run(self) -> List[Node]:
        env = self.env
        typ = self.name
        if typ == "solution" and env.app.config.hide_solutions:
            return []

        serial_no = env.new_serialno()

        if not hasattr(env, "exercise_list"):
            env.exercise_list = {}

        classes, class_name = [typ], self.options.get("class", "")
        if class_name:
            classes.extend(class_name)

        # Have a dummy title text if no title specified, as 'std' domain needs
        # a title to process it as enumerable node.
        if typ == "exercise":
            title_text = f"{self.name.title()} "

            if self.arguments != []:
                title_text = f"({self.arguments[0]})"
        else:
            title_text = f"{self.name.title()} to "
            target_label = self.arguments[0]

        # selecting the type of node
        if typ == "exercise":
            if "nonumber" in self.options:
                node = exercise_unenumerable_node()
            else:
                node = exercise_node()
        else:
            node = solution_node()

        # state parsing
        section = nodes.section(ids=[f"{typ}-content"])
        textnodes, messages = self.state.inline_text(title_text, self.lineno)
        self.state.nested_parse(self.content, self.content_offset, section)

        node += nodes.title(title_text, "", *textnodes)
        node += section

        label = self.options.get("label", "")
        if label:
            self.options["noindex"] = False
        else:
            self.options["noindex"] = True
            label = f"{env.docname}-{typ}-{serial_no}"

        # Duplicate label warning
        if not label == "" and label in env.exercise_list.keys():
            docpath = env.doc2path(env.docname)
            path = docpath[: docpath.rfind(".")]
            other_path = env.doc2path(env.exercise_list[label]["docname"])
            msg = f"duplicate label: {label}; other instance in {other_path}"
            logger.warning(msg, location=path, color="red")
            return []

        self.options["name"] = label

        # Set node attributes
        node["classes"].extend(classes)
        node["ids"].append(label)
        node["label"] = label
        node["docname"] = env.docname
        node["title"] = title_text
        node["type"] = typ
        node["hidden"] = True if "hidden" in self.options else False
        node.document = self.state.document

        if typ == "solution":
            node["target_label"] = target_label

        self.add_name(node)
        env.exercise_list[label] = {
            "type": typ,
            "docname": env.docname,
            "node": node,
            "title": title_text,
            "hidden": node.get("hidden", bool),
        }

        if node.get("hidden", bool):
            return []

        return [node]


class ExerciseDirective(CustomDirective):
    """ A custom exercise directive """

    name = "exercise"
    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {
        "label": directives.unchanged_required,
        "class": directives.class_option,
        "nonumber": directives.flag,
        "hidden": directives.flag,
    }


class SolutionDirective(CustomDirective):
    """ A custom solution directive """

    name = "solution"
    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        "label": directives.unchanged_required,
        "class": directives.class_option,
        "hidden": directives.flag,
    }



# Utility functions
from sphinx.writers.latex import LaTeXTranslator
from docutils import nodes as docutil_nodes


def find_parent(env, node, parent_tag):
    """Find the nearest parent node with the given tagname."""
    while True:
        node = node.parent
        if node is None:
            return None
        # parent should be a document in toc
        if (
            "docname" in node.attributes
            and env.titles[node.attributes["docname"]].astext().lower()
            in node.attributes["names"]
        ):
            return node.attributes["docname"]

    if node.tagname == parent_tag:
        return node.attributes["docname"]

    return None


def get_node_number(self, node, typ) -> str:
    """Get the number for the directive node for HTML."""
    ids = node.attributes.get("ids", [])[0]
    if isinstance(self, LaTeXTranslator):
        docname = find_parent(self.builder.env, node, "section")
    else:
        docname = node.attributes.get("docname", "")
        # Latex does not have builder.fignumbers
    fignumbers = self.builder.env.toc_fignumbers.get(docname, {})
    number = fignumbers.get(typ, {}).get(ids, ())
    return ".".join(map(str, number))


def has_math_child(node):
    """ Check if a parent node as a math child node. """
    for item in node:
        if isinstance(item, docutil_nodes.math):
            return True
    return False


def get_refuri(node):
    """ Check both refuri and refid, to see which one is available. """
    id_ = ""
    if node.get("refuri", ""):
        id_ = node.get("refuri", "")

    if node.get("refid", ""):
        id_ = node.get("refid", "")

    return id_.split("#")[-1]


def list_rindex(li, x) -> int:
    """Getting the last occurence of an item in a list."""
    for i in reversed(range(len(li))):
        if li[i] == x:
            return i
    raise ValueError("{} is not in list".format(x))



# -*- coding: utf-8 -*-
"""
    sphinx_exercise
    ~~~~~~~~~~~~~~~
    This package is a namespace package that contains all extensions
    distributed in the ``sphinx-contrib`` distribution.
    :copyright: Copyright 2007-2009 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""
from pathlib import Path
from typing import Any, Dict, Set, Union, cast
from sphinx.config import Config
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.domains.std import StandardDomain
from sphinx.addnodes import number_reference
from docutils.nodes import Node
from docutils import nodes as docutil_nodes
from sphinx.util import logging
from sphinx.util.fileutil import copy_asset
from sphinx.builders.latex import LaTeXBuilder
from sphinx.transforms.post_transforms import SphinxPostTransform

from .directive import ExerciseDirective, SolutionDirective
from .nodes import (
    exercise_node,
    exercise_unenumerable_node,
    solution_node,
    visit_enumerable_node,
    depart_enumerable_node,
    visit_exercise_unenumerable_node,
    depart_exercise_unenumerable_node,
    visit_solution_node,
    depart_solution_node,
    is_solution_node,
    is_exercise_node,
    is_exercise_unenumerable_node,
    is_extension_node,
    NODE_TYPES,
)
from .utils import get_node_number, get_refuri, has_math_child, find_parent

logger = logging.getLogger(__name__)

# Variables
SOLUTION_PLACEHOLDER = "Solution to "
MATH_PLACEHOLDER = ":math:"


def purge_exercises(app: Sphinx, env: BuildEnvironment, docname: str) -> None:
    if not hasattr(env, "exercise_list"):
        return

    # Override env.exercise_list
    env.exercise_list = {
        exercise: env.exercise_list[exercise]
        for exercise in env.exercise_list.keys()
        if env.exercise_list[exercise]["docname"] != docname
    }


def merge_exercises(
    app: Sphinx, env: BuildEnvironment, docnames: Set[str], other: BuildEnvironment
) -> None:
    if not hasattr(env, "exercise_list"):
        env.exercise_list = {}

    # Merge env stored data
    if hasattr(other, "exercise_list"):
        env.exercise_list = {**env.exercise_list, **other.exercise_list}


def init_numfig(app: Sphinx, config: Config) -> None:
    """Initialize exercise numfig format."""
    config["numfig"] = True
    numfig_format = {}
    for typ in NODE_TYPES.keys():
        numfig_format[typ] = typ.title() + " %s"
    numfig_format.update(config.numfig_format)
    config.numfig_format = numfig_format


def copy_asset_files(app: Sphinx, exc: Union[bool, Exception]):
    static_path = Path(__file__).parent.joinpath("_static", "exercise.css").absolute()
    asset_files = [str(static_path)]

    if exc is None:
        for path in asset_files:
            copy_asset(path, str(Path(app.outdir).joinpath("_static").absolute()))


def doctree_read(app: Sphinx, document: Node) -> None:
    domain = cast(StandardDomain, app.env.get_domain("std"))

    # Traverse extension nodes
    for node in document.traverse():
        docname, labelid, sectname = "", "", ""

        if is_extension_node(node):
            name = node.get("names", [])[0]
            labelid = document.nameids[name]
            docname = app.env.docname

            # If solution node
            if is_solution_node(node):
                sectname = SOLUTION_PLACEHOLDER
            else:
                # If other node, simply add :math: to title
                # to allow for easy parsing in ref_node
                for item in node[0]:
                    if isinstance(item, docutil_nodes.math):
                        sectname += f"{MATH_PLACEHOLDER}`{item.astext()}` "
                        continue
                    sectname += f"{item.astext()} "

                # Lastly, remove parans from title
                _r, _l = sectname.rfind(")"), sectname.find("(") + 1
                sectname = sectname[_l:_r].strip()

            domain.anonlabels[name] = docname, labelid
            domain.labels[name] = docname, labelid, sectname


def update_title(title):
    """
    Does necessary formatting to the title node, and wraps it with an inline node.
    """
    inline = docutil_nodes.inline()

    if len(title) == 1 and isinstance(title[0], docutil_nodes.Text):
        _ = title[0][0].replace("(", "").replace(")", "")
        inline += docutil_nodes.Text(_, _)
    else:
        for ii in range(len(title)):
            item = title[ii]

            if ii == 0 and isinstance(item, docutil_nodes.Text):
                _ = item.replace("exercise", "").replace("(", "").lstrip()
                title.replace(title[ii], docutil_nodes.Text(_, _))
            elif ii == len(title) - 1 and isinstance(item, docutil_nodes.Text):
                _ = item.replace(")", "").rstrip()
                if _:
                    title.replace(title[ii], docutil_nodes.Text(_, _))
                else:
                    continue
            inline += title[ii]

    return inline


def process_math_placeholder(node, update_title, source_node):
    """Convert the placeholder math text to a math node."""
    if MATH_PLACEHOLDER in node.astext():
        title = update_title(source_node[0])
        return node.replace(node[0], title)


def process_reference(self, node, default_title=""):
    """
    Processing reference nodes in the document to facilitate the design and the
    functionality requirements.
    """
    label = get_refuri(node)
    if hasattr(self.env, "exercise_list") and label in self.env.exercise_list:
        source_node = self.env.exercise_list[label].get("node")
        # if reference source is a solution node
        if is_solution_node(source_node):
            target_label = source_node.attributes.get("target_label", "")
            if node.astext().strip() == "Solution to":
                default_title = node.astext()
        else:
            target_label = source_node.attributes.get("label", "")
        target_attr = self.env.exercise_list[target_label]
        target_node = target_attr.get("node", Node)
        # if reference target is exercise node
        if is_exercise_node(target_node):
            if default_title:
                number = get_node_number(self.app, target_node, "exercise")
                node.insert(len(node[0]), docutil_nodes.Text(" Exercise " + number))
                return
            else:
                node = process_math_placeholder(node, update_title, source_node)
        # if reference target is an exercise unenumerable node
        if is_exercise_unenumerable_node(target_node):
            if default_title:
                if target_attr.get("title"):
                    if has_math_child(target_node[0]):
                        title = update_title(target_node[0])
                        title.insert(
                            0, docutil_nodes.Text(default_title, default_title)
                        )
                        node.replace(node[0], title)
                    else:
                        text = target_attr.get("title", "")
                        if text[0] == "(" and text[-1] == ")":
                            text = text[1:-1]
                        node[0].insert(len(node[0]), docutil_nodes.Text(text, text))
            else:
                node = process_math_placeholder(node, update_title, source_node)


class ReferenceTransform(SphinxPostTransform):
    default_priority = 998  # should be processed before processing solution nodes

    def run(self):

        for node in self.document.traverse(docutil_nodes.reference):
            process_reference(self, node)


class SolutionTransorm(SphinxPostTransform):
    default_priority = 999  # should be after processing reference nodes

    def run(self):

        for node in self.document.traverse(solution_node):
            target_labelid = node.get("target_label", "")
            try:
                target_attr = self.env.exercise_list[target_labelid]
            except Exception:
                # target_labelid not found
                if isinstance(self.app.builder, LaTeXBuilder):
                    docname = find_parent(self.app.builder.env, node, "section")
                else:
                    docname = self.app.builder.current_docname
                docpath = self.env.doc2path(docname)
                path = docpath[: docpath.rfind(".")]
                msg = f"undefined label: {target_labelid}"
                logger.warning(msg, location=path, color="red")
                return

            # Create a reference
            refuri = self.app.builder.get_relative_uri(
                self.env.docname, target_attr.get("docname", "")
            )
            refuri += "#" + target_labelid

            # create a text for the reference node
            title = node.attributes["title"]
            newtitle = docutil_nodes.inline("", docutil_nodes.Text(title))
            reference = docutil_nodes.reference(
                "",
                "",
                internal=True,
                refuri=refuri,
                anchorname="",
                *[newtitle],
            )
            process_reference(self, reference, SOLUTION_PLACEHOLDER)
            newnode = docutil_nodes.title("")
            newnode.append(reference)
            node[0].replace_self(newnode)
            # update node
            self.env.exercise_list[node.get("label", "")]["node"] = node


class NumberReferenceTransform(SphinxPostTransform):
    default_priority = 1000

    def run(self):

        for node in self.document.traverse(number_reference):
            labelid = get_refuri(node)

            # If extension directive referenced
            if labelid in self.env.exercise_list:
                source_attr = self.env.exercise_list[labelid]
                source_node = source_attr.get("node", Node)
                node_title = node.get("title", "")

                # processing for nodes which have
                if "{name}" in node_title and has_math_child(source_node[0]):
                    newtitle = docutil_nodes.inline()
                    for item in node_title.split():
                        if item == "{name}":
                            # use extend instead?
                            for _ in update_title(source_node[0]):
                                newtitle += _
                        elif item == "{number}":
                            source_type = source_node.attributes["type"]
                            source_number = get_node_number(
                                self.app, source_node, source_type
                            )
                            source_num = ".".join(map(str, source_number))
                            newtitle += docutil_nodes.Text(source_num, source_num)
                        else:
                            newtitle += docutil_nodes.Text(item, item)
                        newtitle += docutil_nodes.Text(" ", " ")

                    if newtitle[len(newtitle) - 1].astext() == " ":
                        newtitle.pop()
                    node.replace(node[0], newtitle)


def setup(app: Sphinx) -> Dict[str, Any]:

    app.add_config_value("hide_solutions", False, "env")

    app.add_css_file("exercise.css")
    app.connect("config-inited", init_numfig)  # event order - 1
    app.connect("env-purge-doc", purge_exercises)  # event order - 5 per file
    app.connect("doctree-read", doctree_read)  # event order - 8
    app.connect("env-merge-info", merge_exercises)  # event order - 9
    app.connect("build-finished", copy_asset_files)  # event order - 16

    app.add_enumerable_node(
        exercise_node,
        "exercise",
        None,
        singlehtml=(visit_enumerable_node, depart_enumerable_node),
        html=(visit_enumerable_node, depart_enumerable_node),
        latex=(visit_enumerable_node, depart_enumerable_node),
    )

    app.add_node(
        exercise_unenumerable_node,
        singlehtml=(
            visit_exercise_unenumerable_node,
            depart_exercise_unenumerable_node,
        ),
        html=(visit_exercise_unenumerable_node, depart_exercise_unenumerable_node),
        latex=(visit_exercise_unenumerable_node, depart_exercise_unenumerable_node),
    )

    app.add_enumerable_node(
        solution_node,
        "solution",
        None,
        singlehtml=(visit_solution_node, depart_solution_node),
        html=(visit_solution_node, depart_solution_node),
        latex=(visit_solution_node, depart_solution_node),
    )

    app.add_directive("exercise", ExerciseDirective)
    app.add_directive("solution", SolutionDirective)

    app.add_post_transform(ReferenceTransform)
    app.add_post_transform(SolutionTransorm)
    app.add_post_transform(NumberReferenceTransform)

    return {
        "version": "builtin",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
sphinx_exercise.nodes
~~~~~~~~~~~~~~~~~~~~~

Enumerable and unenumerable nodes

:copyright: Copyright 2020 by the QuantEcon team, see AUTHORS
:licences: see LICENSE for details
"""
from docutils.nodes import Node
from sphinx.util import logging
from docutils import nodes as docutil_nodes
from sphinx.writers.latex import LaTeXTranslator
from .utils import get_node_number, find_parent, list_rindex

logger = logging.getLogger(__name__)

CR = "\n"
latex_admonition_start = CR + "\\begin{sphinxadmonition}{note}"
latex_admonition_end = "\\end{sphinxadmonition}" + CR


class exercise_node(docutil_nodes.Admonition, docutil_nodes.Element):
    pass


class solution_node(docutil_nodes.Admonition, docutil_nodes.Element):
    pass


class exercise_unenumerable_node(docutil_nodes.Admonition, docutil_nodes.Element):
    pass


def _visit_nodes_latex(self, node, find_parent):
    """ Function to handle visit_node for latex. """
    docname = find_parent(self.builder.env, node, "section")
    self.body.append(
        "\\phantomsection \\label{" + f"{docname}:{node.attributes['label']}" + "}"
    )
    self.body.append(latex_admonition_start)


def _depart_nodes_latex(self, node, title, pop_index=False):
    """ Function to handle depart_node for latex. """
    idx = list_rindex(self.body, latex_admonition_start) + 2
    if pop_index:
        self.body.pop(idx)
    self.body.insert(idx, title)
    self.body.append(latex_admonition_end)


def _remove_placeholder_title_exercise(typ, node):
    """ Removing the exercise placeholder we put in title earlier."""
    for title in node.traverse(docutil_nodes.title):
        if typ.title() in title.astext():
            title[0] = docutil_nodes.Text("")


def visit_enumerable_node(self, node: Node) -> None:
    typ = node.attributes.get("type", "")
    if isinstance(self, LaTeXTranslator):
        _remove_placeholder_title_exercise(typ, node)
        _visit_nodes_latex(self, node, find_parent)
    else:
        _remove_placeholder_title_exercise(typ, node)
        self.body.append(self.starttag(node, "div", CLASS="admonition"))


def depart_enumerable_node(self, node: Node) -> None:
    typ = node.attributes.get("type", "")
    if isinstance(self, LaTeXTranslator):
        number = get_node_number(self, node, typ)
        _depart_nodes_latex(self, node, f"{typ.title()} {number} ")
    else:
        number = get_node_number(self, node, typ)
        idx = list_rindex(self.body, f"{typ.title()} {number} ")
        self.body[idx] = f"{typ.title()} {number} "
        self.body.append("</div>")


def visit_exercise_unenumerable_node(self, node: Node) -> None:
    typ = node.attributes.get("type", "")
    if isinstance(self, LaTeXTranslator):
        _remove_placeholder_title_exercise(typ, node)
        _visit_nodes_latex(self, node, find_parent)
    else:
        _remove_placeholder_title_exercise(typ, node)
        self.body.append(self.starttag(node, "div", CLASS="admonition"))


def depart_exercise_unenumerable_node(self, node: Node) -> None:
    typ = node.attributes.get("type", "")
    if isinstance(self, LaTeXTranslator):
        _depart_nodes_latex(self, node, f"{typ.title()} ")
    else:
        idx = list_rindex(self.body, '<p class="admonition-title">') + 1
        element = f"<span>{typ.title()} </span>"
        self.body.insert(idx, element)
        self.body.append("</div>")


def visit_solution_node(self, node: Node) -> None:
    if isinstance(self, LaTeXTranslator):
        _visit_nodes_latex(self, node, find_parent)
    else:
        self.body.append(self.starttag(node, "div", CLASS="admonition"))


def depart_solution_node(self, node: Node) -> None:
    typ = node.attributes.get("type", "")
    if isinstance(self, LaTeXTranslator):
        _depart_nodes_latex(self, node, f"{typ.title()} to ", True)
    else:
        number = get_node_number(self, node, typ)
        idx = list_rindex(self.body, f"{typ.title()} {number} ")
        self.body.pop(idx)
        self.body.append("</div>")


def is_exercise_node(node):
    return isinstance(node, exercise_node)


def is_exercise_unenumerable_node(node):
    return isinstance(node, exercise_unenumerable_node)


def is_solution_node(node):
    return isinstance(node, solution_node)


def is_extension_node(node):
    return (
        is_exercise_node(node)
        or is_exercise_unenumerable_node(node)
        or is_solution_node(node)
    )


def rreplace(s, old, new, occurrence):
    # taken from https://stackoverflow.com/a/2556252
    li = s.rsplit(old, occurrence)
    return new.join(li)


NODE_TYPES = {
    "exercise": {"node": exercise_node, "type": "exercise"},
    "solution": {"node": solution_node, "type": "solution"},
}



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

VERSION = "v0.2.1"

LONG_DESCRIPTION = """
This package contains a [Sphinx](http://www.sphinx-doc.org/en/master/) extension
for producing exercise and solution directives.

This project is maintained and supported by [najuzilu](https://github.com/najuzilu).
"""

SHORT_DESCRIPTION = "A Sphinx extension for producing exercises and solutions."

BASE_URL = "https://github.com/executablebooks/sphinx-exercise"
URL = f"{BASE_URL}/archive/{VERSION}.tar.gz"

# Define all extras
extras = {
    "code_style": ["flake8<3.8.0,>=3.7.0", "black", "pre-commit==1.17.0"],
    "testing": [
        "coverage",
        "pytest>=3.6,<4",
        "pytest-cov",
        "pytest-regressions",
        "beautifulsoup4",
        "myst-nb",
        "texsoup",
    ],
    "rtd": [
        "sphinx>=3.0",
        "sphinx-book-theme",
        "myst-nb",
    ],
}

extras["all"] = set(ii for jj in extras.values() for ii in jj)

setup(
    name="sphinx-exercise",
    version=VERSION,
    python_requires=">=3.6",
    author="QuantEcon",
    author_email="admin@quantecon.org",
    url=BASE_URL,
    download_url=URL,
    project_urls={
        "Source": BASE_URL,
        "Tracker": f"{BASE_URL}/issues",
    },
    description=SHORT_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="BSD",
    packages=find_packages(),
    install_requires=["docutils>=0.15", "sphinx", "sphinx-book-theme"],
    extras_require=extras,
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Plugins",
        "Environment :: Web Environment",
        "Framework :: Sphinx :: Extension",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python",
        "Topic :: Documentation :: Sphinx",
        "Topic :: Documentation",
        "Topic :: Software Development :: Documentation",
        "Topic :: Text Processing",
        "Topic :: Utilities",
    ],
)
